﻿using ChunSun.NginxServer.DaemonService.Utility;
using System;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Management.Automation;
using System.ServiceModel;
using System.Text;

namespace ChunSun.NginxServer.DaemonService.Impl
{
    [ServiceContract]
    public interface IEntranceService
    {
        /// <summary>
        /// 绑定域名。
        /// </summary>
        /// <param name="domains">域名数组。</param>
        /// <param name="url">被代理的url。</param>
        [OperationContract]
        void BindingDomains(string[] domains, string url);

        /// <summary>
        /// 删除域名绑定。
        /// </summary>
        /// <param name="domains">域名数组。</param>
        [OperationContract]
        void RemoveDomainBindings(string[] domains);
    }

    internal sealed class EntranceService : IEntranceService
    {
        #region Field

        private readonly string _nginxPath;
        private readonly PersistenceList<DomainModel> _list;

        //Nginx Reload 运行模式（PowerShell、CMD）
        private readonly string _reloadRunMode;

        private static readonly object SyncLock = new object();

        #endregion Field

        #region Property

        //        public ILogger Logger { get; set; }

        #endregion Property

        #region Constructor

        public EntranceService()
        {
            //            Logger = WorkContextScope.Resolve<ILoggerFactory>().CreateLogger(GetType());

            //根据配置得到Nginx路径。
            _nginxPath = ConfigurationManager.AppSettings["NginxPath"];
            //根据配置得到Reload运行模式。
            _reloadRunMode = ConfigurationManager.AppSettings["RunMode"];

            _list = new PersistenceList<DomainModel>(MapPath("domains.xml"));
        }

        #endregion Constructor

        #region Implementation of IEntranceService

        /// <summary>
        /// 绑定域名。
        /// </summary>
        /// <param name="domains">域名数组。</param>
        /// <param name="url">被代理的url。</param>
        public void BindingDomains(string[] domains, string url)
        {
            //参数验证、优化。
            if (domains == null)
                throw new ArgumentNullException("domains");
            if (string.IsNullOrWhiteSpace(url))
                throw new ArgumentNullException("url");

            domains = domains.Select(i => i.Trim().ToLower()).ToArray();
            url = url.Trim().ToLower();

            if (domains.Count(IsDomain) <= 0)
                throw new Exception("在需要绑定的域名中发现非法数据。");
            if (!IsUrl(url))
                throw new Exception("被代理的地址不是一个合法的Url地址。");

            //如果需要绑定的域名已经存在则先删除。
            InternalRemoveDomainBindings(domains);

            var domainModel = _list.SingleOrDefault(i => string.Equals(i.Url, url, StringComparison.OrdinalIgnoreCase));
            if (domainModel == null)
            {
                domainModel = new DomainModel
                {
                    Url = url
                };
                _list.Add(domainModel);
            }
            domainModel.Domains = domains.Concat(domainModel.Domains).Distinct().ToArray();

            Refresh();
        }

        /// <summary>
        /// 删除域名绑定。
        /// </summary>
        /// <param name="domains">域名数组。</param>
        public void RemoveDomainBindings(string[] domains)
        {
            if (domains.Count(IsDomain) <= 0)
                throw new Exception("在需要解除绑定的域名中发现非法数据。");

            InternalRemoveDomainBindings(domains);
            Refresh();
        }

        #endregion Implementation of IEntranceService

        #region Private Method

        private void InternalRemoveDomainBindings(string[] domains)
        {
            var result = _list.Where(i => i.Domains.Any(domains.Contains)).ToArray();
            foreach (var domainModel in result)
            {
                domainModel.Domains = domainModel.Domains.Except(domains).ToArray();
                //没有任何域名时删除该记录。
                if (!domainModel.Domains.Any())
                    _list.Remove(domainModel);
            }
        }

        private void Refresh()
        {
            for (var i = 0; i < 3; i++)
            {
                lock (SyncLock)
                {
                    try
                    {
                        //同步Nginx配置文件。
                        SyncConfig();
                        //持久化本地配置文件。
                        _list.Save();
                        //通知Nginx重新载入配置。
                        ReloadNginx();
                        break;
                    }
                    catch (Exception exception)
                    {
                        //                        Logger.Fatal(exception, "刷新状态时候发生了异常。");
#if DEBUG
                        throw;
#endif
                    }
                }
            }
        }

        private void SyncConfig()
        {
            var chunSunConfig = Path.Combine(_nginxPath, "conf", "custom.conf");
            var builder = new StringBuilder();

            //筛选合法的域名模型。
            var domains = _list.Where(i => i.Domains.Count(IsDomain) > 0 && IsUrl(i.Url));
            foreach (var domainModel in domains)
            {
                builder.AppendLine("    server {");
                builder.AppendLine("	listen       80;");

                builder.Append("	server_name ");
                builder.Append(string.Join(" ", domainModel.Domains));
                builder.AppendLine(";");

                builder.AppendLine("        location / {");

                builder.Append("            proxy_pass ");
                builder.Append(domainModel.Url);
                builder.AppendLine(";");

                builder.AppendLine("        }");
                builder.AppendLine("    }");
            }
            File.WriteAllText(chunSunConfig, builder.ToString());
        }

        private void ReloadNginx()
        {
            switch (_reloadRunMode == null ? null : _reloadRunMode.ToLower())
            {
                case "powershell":
                    using (var powerShell = PowerShell.Create())
                    {
                        powerShell.AddCommand("cd").AddArgument(_nginxPath).Invoke();
                        powerShell.Commands.Clear();
                        powerShell.AddCommand(".\\nginx").AddArgument("-s").AddArgument("reload");
                        powerShell.Invoke();
                    }
                    break;

                default:
                    var tempReloadPath = GetTempReloadPath();

                    using (var process = new Process
                    {
                        StartInfo =
                        {
                            FileName = tempReloadPath,
                            UseShellExecute = false
                        }
                    })
                    {
                        process.Start();
                    }
                    break;
            }
        }

        private string GetTempReloadPath()
        {
            var tempReloadPath = MapPath("TempReload.bat");

            if (File.Exists(tempReloadPath))
                File.Delete(tempReloadPath);

            var directory = new DirectoryInfo(_nginxPath);
            Func<DirectoryInfo, string> getRoot = null;
            getRoot = d => d.Parent != null ? getRoot(d.Parent) : d.Name.TrimEnd('\\', '/');

            var root = getRoot(directory);
            var path = directory.FullName.Substring(root.Length + 1);

            var reloadPath = MapPath("Reload.bat");
            File.WriteAllText(tempReloadPath, File.ReadAllText(reloadPath).Replace("{S}", root).Replace("{NginxPath}", path));

            return tempReloadPath;
        }

        private static bool IsDomain(string domain)
        {
            return !string.IsNullOrWhiteSpace(domain) && IsUrl(string.Format("http://{0}/", domain));
        }

        private static bool IsUrl(string url)
        {
            Uri uri;
            return !string.IsNullOrWhiteSpace(url) && Uri.TryCreate(url, UriKind.Absolute, out uri);
        }

        private static string MapPath(string path)
        {
            return Path.Combine(AppDomain.CurrentDomain.BaseDirectory, path);
        }

        #endregion Private Method
    }

    public class DomainModel
    {
        public DomainModel()
        {
            Domains = new string[0];
        }

        public string[] Domains { get; set; }

        public string Url { get; set; }
    }
}